(* type des termes utilisant le type 'a pour reprsenter les
   constructeurs.  Attention : On utilise l'galit physique pour
   comparer les constructeurs ! *)

type 'a term =
  Fun of ('a * 'a term list)
| Var of int
;;

(* type des substitutions. On supposera les variables en ordre
   strictement croissant afin d'avoir une composition plus rapide. *)

type 'a subst == (int * 'a term) list;;

(* la fonction assoc tirant partie de l'ordre *)

let ord_assoc n l = fn l
  where rec fn = function
  [] -> raise Not_found
| (n',t)::l ->
    if n' = n then t
    else if n < n' then raise Not_found
    else fn l
;;

(* Une fonction merge utilisant l'ordre et ne gardant que la premire
   occurence d'une variable n (important pour la composition des
   substitutions). *)

let rec ord_merge = fun
  [] l -> l
| l [] -> l
| ((n,_ as c)::l as l0) ((n',_ as c')::l' as l0') ->
  if n = n' then c::(ord_merge l l')
  else if n < n' then c::(ord_merge l l0')
  else c'::(ord_merge l0 l')
;;

(* Application d'une substitution  un terme *)

let subst sigma t =fn t 
  where rec fn = function
    Var n as t -> (try ord_assoc n sigma with Not_found -> t)
  | Fun (i,l)  -> Fun (i, map fn l)
;;

(* Composition de deux substitutions (Attention ce n'est pas aussi simple
   qu'il y parait, cf les dtails de ord_merge) *)

let compose sigma' sigma = 
  ord_merge (map (function n,t -> n,subst sigma' t) sigma) sigma'
;;

(* Test d'occurrence *)

let occur n t = fn t
  where rec fn = function
    Var n' -> n = n'
  | Fun(_,l) -> exists fn l
;;

(* Unification d'une liste d'quations (paire de termes). La
   substitution partiellement construite est conserv en premier
   argument.  *)

exception Clash;;
exception Occur_check;;

let rec unifs sigma = function
  [] -> sigma
| (Fun(i,l),Fun(i',l'))::eqs ->
    if i == i' then unifs sigma ((combine (l,l'))@eqs)
    else raise Clash
| (Var x,t)::eqs ->
    if t = Var x then unifs sigma eqs else begin
      if occur x t then raise Occur_check;
      let s = [x,t] in
      unifs (compose s sigma) 
            (map (function t,t' -> subst s t, subst s t') eqs)
    end
| (t, Var x)::eqs ->
    if t = Var x then unifs sigma eqs else begin
      if occur x t then raise Occur_check;
      let s = [x,t] in
      unifs (compose s sigma) 
            (map (function t,t' -> subst s t, subst s t') eqs)
    end
;;

(* Unification de deux termes *)

let unif u v = unifs [] [u,v];;

(* examples *)

let f u = Fun("f", [u]);;
let g u = Fun("g", [u]);;
let h u v = Fun("h", [u;v]);;
let x = Var 0 and y = Var 1 and z = Var 2 and s = Var 3 and t = Var 4;;

unif (f x) (g x);;
unif (f (f x)) (f y);;
unif (f (f x)) (f x);;
unif (h x (h x x)) (h (h z z) y);;
unif (h x (h x x)) (h (h z z) x);;

(* Une srie exponentielle *)
unif (h x (h y y)) (h (h y y) x);;
unif (h x (h y (h z z))) (h (h (h z z) y) x);;
unif (h x (h y (h z (h s s)))) (h (h (h (h s s) z) y) x);;
unif (h x (h y (h z (h s (h t t))))) (h (h (h (h s (h t t)) z) y) x);;
