(* Hash tables *)


(* We do dynamic hashing, and we double the size of the table when
   buckets become too long, but without re-hashing the elements. *)

type ('a, 'b) t =
  { mutable max_len: int;                    (* max length of a bucket *)
    mutable data: ('a, 'b) bucketlist vect;
    hashfun : ('a -> int) choice;
    equalfun : ('a -> 'a -> bool) choice } (* the buckets *)

and ('a, 'b) bucketlist =
    Empty
  | Cons of ('a, 'b) bucket

and ('a, 'b) bucket =
  {Hash : int; Key : 'a; mutable Val : 'b; mutable Next : ('a, 'b) bucketlist}
;;

let bucket_info buck = buck.Val;;
let bucket_key buck = buck.Key;;

let new initial_size hashfun equalfun =
  { max_len = 1; data = make_vect initial_size Empty;
    hashfun = hashfun; equalfun = equalfun }
;;

let hash t key =
  match t.hashfun with
    Standard -> hashtbl__hash_param 10 100 key
  | Custom f -> f key
;;

let equal t key key' =
  match t.equalfun with
    Standard -> key = key'
  | Custom f -> f key key'
;;

let clear h =
  for i = 0 to vect_length h.data - 1 do
    h.data.(i) <- Empty
  done
;;

let resize h =
  let n = vect_length h.data in
  let nn = n+n in
  let newdata = make_vect nn Empty in
    let rec fn = function
      Empty -> ()
    | Cons ({Hash = hn; Next = rest} as bucket) -> 
        fn rest;
        let r = newdata.(hn mod nn) in
        rest <- r; newdata.(hn mod nn) <- Cons bucket
    in do_vect fn h.data; 
    h.data <- newdata;
    ()
;;

let rec bucket_too_long hl n bucket =
  if n < 0 then true else
    match bucket with
      Empty -> false
    | Cons{Hash = hash; Next = rest} -> 
        if mem hash hl then
          bucket_too_long hl n rest
        else 
          bucket_too_long (hash::hl) n rest
;;

let add h key info =
  let hn = hash h key in
  let i = hn mod (vect_length h.data) in
  let bucket = Cons{Hash = hn; Key=key; Val=info; Next=h.data.(i)} in
    h.data.(i) <- bucket;
    if bucket_too_long [] h.max_len bucket then resize h
;;

let add_bucket h key info =
  let hn = hash h key in
  let i = hn mod (vect_length h.data) in
  let bucket = {Hash = hn; Key=key; Val=info; Next=h.data.(i)} in
    h.data.(i) <- Cons bucket;
    if bucket_too_long [] h.max_len h.data.(i) then resize h;
    bucket
;;

let remove h key =
  let i = (hash h key) mod (vect_length h.data) in
    
  match h.data.(i) with
    Empty -> raise Not_found
  | Cons{Key = k1; Next = rest1} ->
      if equal h key k1 then h.data.(i) <- rest1 else
      match rest1 with
        Empty -> raise Not_found
      | Cons{Key = k2; Next = rest2} ->
          if equal h key k2 then rest1 <- rest2 else
          match rest2 with
            Empty -> raise Not_found
          | Cons({Key = k3; Next = rest3} as last) ->
              if equal h key k3 then rest2 <- rest3 else begin
                let rec rm {Next = pr} = function
                    Empty ->
                      raise Not_found
                  | Cons({Key = k; Next = rest} as last) ->
                      if equal h key k then pr <- rest else rm last rest
                in rm last rest3
              end
;;

let fast_remove h ({Hash = hn} as bucket) =
  let i = hn mod (vect_length h.data) in
    
  match h.data.(i) with
    Empty -> raise Not_found
  | Cons({Next = rest1} as buck) ->
      if buck == bucket then h.data.(i) <- rest1 else
      match rest1 with
        Empty -> raise Not_found
      | Cons({Next = rest2} as buck)->
          if buck == bucket then rest1 <- rest2 else
          match rest2 with
            Empty -> raise Not_found
          | Cons({Next = rest3} as buck) ->
              if buck == bucket then rest2 <- rest3 else begin
                let rec rm {Next = pr} = function
                    Empty ->
                      raise Not_found
                  | Cons({Next = rest} as buck) ->
                      if buck == bucket then pr <- rest else rm buck rest
                in rm buck rest3
              end
;;


let find h key =
  match h.data.((hash h key) mod (vect_length h.data)) with
    Empty -> raise Not_found
  | Cons{Key = k1; Val = d1; Next = rest1} ->
      if equal h key k1 then d1 else
      match rest1 with
        Empty -> raise Not_found
      | Cons{Key = k2; Val = d2; Next = rest2} ->
          if equal h key k2 then d2 else
          match rest2 with
            Empty -> raise Not_found
          | Cons{Key = k3; Val = d3; Next = rest3} ->
              if equal h key k3 then d3 else begin
                let rec find = function
                    Empty ->
                      raise Not_found
                  | Cons{Key = k; Val = d; Next = rest} ->
                      if equal h key k then d else find rest
                in find rest3
              end
;;

let add_find h key ft =
  let hn = hash h key in
  let i = hn mod (vect_length h.data) in  
  find h.data.(i)
  where rec find = function
    Empty ->  
      let info = ft key in
      let bucket = {Hash = hn; Key=key; Val=info; Next=h.data.(i)} in
      h.data.(i) <- Cons bucket;
      if bucket_too_long [] h.max_len h.data.(i) then resize h;
      info
  | Cons{Key = k; Val = d; Next = rest} ->
      if equal h key k then d else find rest
;;

let find_all h key =
  let rec find_in_bucket = function
    Empty ->
      []
  | Cons{Key = k; Val = d; Next = rest} ->
      if equal h k key then d :: find_in_bucket rest 
                       else find_in_bucket rest in
  find_in_bucket h.data.((hash h key) mod (vect_length h.data))
;;

let do_table f h =
  let len = vect_length h.data in
  for i = 0 to vect_length h.data - 1 do
    let rec do_bucket = function
        Empty ->
          ()
      | Cons{Key = k; Val = d; Next = rest} ->
          f k d; do_bucket rest in
    do_bucket h.data.(i)
  done
;;

let it_table f a h =
  let len = vect_length h.data in
  let res = ref a in
  for i = 0 to vect_length h.data - 1 do
    let rec do_bucket = function
        Empty ->
          ()
      | Cons{Key = k; Val = d; Next = rest} ->
          res := f !res k d; do_bucket rest in
    do_bucket h.data.(i)
  done;
  !res
;;

