(* Hash tables and hash functions *)

(* Hash tables are hashed association tables, with in-place modification. *)

type 'a choice = 
  Standard
| Custom of 'a
;;

type ('a, 'b) t;;
        (* The type of hash tables from type ['a] to type ['b]. *)
type ('a, 'b) bucket;;

value new : int -> ('a -> int) choice -> ('a -> 'a -> bool) choice -> ('a,'b) t
        (* [new n] creates a new, empty hash table, with initial size [n].
           The table grows as needed, so [n] is just an initial guess.
           Better results are said to be achieved when [n] is a prime
           number. *)

  and clear : ('a, 'b) t -> unit
        (* Empty a hash table. *)

  and add : ('a, 'b) t -> 'a -> 'b -> unit
        (* [add tbl x y] adds a binding of [x] to [y] in table [tbl].
           Previous bindings for [x] are not removed, but simply
           hidden. That is, after performing [remove tbl x], the previous
           binding for [x], if any, is restored.
           (This is the semantics of association lists.) *)

  and add_bucket : ('a, 'b) t -> 'a -> 'b -> ('a, 'b) bucket
        (* same has add but send back the new bucket *)

  and find : ('a, 'b) t -> 'a -> 'b
        (* [find tbl x] returns the current binding of [x] in [tbl],
           or raises [Not_found] if no such binding exists. *)

  and find_all : ('a, 'b) t -> 'a -> 'b list
        (* [find_all tbl x] returns the list of all data associated with [x]
           in [tbl]. The current binding is returned first, then the previous
           bindings, in reverse order of introduction in the table. *)

  and remove : ('a, 'b) t -> 'a -> unit
        (* [remove tbl x] removes the current binding of [x] in [tbl],
           restoring the previous binding if it exists.
           It does nothing if [x] is not bound in [tbl]. *)

  and fast_remove : ('a, 'b) t -> ('a, 'b) bucket -> unit
        (* [remove tbl x] removes the current binding of [x] in [tbl],
           restoring the previous binding if it exists.
           It does nothing if [x] is not bound in [tbl]. *)

  and do_table : ('a -> 'b -> 'c) -> ('a, 'b) t -> unit
        (* [do_table f tbl] applies [f] to all bindings in table [tbl],
	   discarding all the results.
           [f] receives the key as first argument, and the associated value
           as second argument. The order in which the bindings are passed to
           [f] is unpredictable. Each binding is presented one time to [f]. *)

  and it_table : ('c -> 'a -> 'b -> 'c) -> 'c -> ('a, 'b) t -> 'c
        (* [do_table f a tbl] is similar to it_list2 f a l1 l2 where 
           l1 is the list of all key in tbl and l2 the list of all data.
           The order in which bindings are passed to [f] is unpredictable. *)

  and add_find : ('a, 'b) t -> 'a -> ('a -> 'b) -> 'b 

  and bucket_info : ('a, 'b) bucket -> 'b

  and bucket_key : ('a, 'b) bucket -> 'a
;;


