(* on a besoin de l'unification linaire .... *)
#open "unif";;

(* les trois stratgies implmentes *)
type strategie = Binary | Hyper | NegHyper;;

(* type des termes (sous forme d'arbre et nom sous forme de graphe
   comme pour l'unification linaire), utilis uniquement avant la
   mise sous forme de clauses *)
type term =
  Fun of (cst_info * term list)
| BVar of int (* indice de De Bruijn de la variable *) 
;;

(* type des prdicats du premier ordre *)
type prop = 
  Vrai | Faux
| Atom of term
| Et of prop * prop
| Ou of prop * prop
| Im of prop * prop
| Eq of prop * prop
| Non of prop
| Al of prop
| Ex of prop
;;

(* type des informations ajoutes aux clauses pour optimiser la subsumption 
     nb de lit. pos., nb de lit. neg.,
     longueur total des lit. pos., 
     max depth dans lit. pos., min depth dans lit pos.,
     idem des trois derniers pour lit neg *)
type info ==
  int * int * int * int * int * int * int * int
;;

(* type des clauses:
     numero de la clause,
     True pour les clauses initiale,
     info (cf ci-dessus),
     liste des lit. pos., listes lit. ng *)
type clause ==
  int ref * bool * info * Term list * Term list
;;

(* Gnration de nouvelles constantes de fonctions *)
let count_cst = ref 0;;
(* Cette fonction est appel par la skolmisation uniquement, l'arit et
   le nom n'ont donc pas d'importance. *)
let new_cst () = 
  let r = !count_cst in count_cst:=r+1; 
  {Cst_name = "!"^(string_of_int r); Cst_arity = 0; Cst_type = Fun_sym};;

(* Remise  zero des compteurs *)
let reset_ref () = 
  count_var := 0; count_cst := 0;;

(* n-ime lt. d'une liste *)
let rec nth = fun
  0 (a::_) -> a
| n (_::l) -> nth (n-1) l
| _ _ -> raise (Invalid_argument "nth")
;;

(* deux exceptions *)
exception Fail;;
exception Proved;;

(* transforme une formule en une liste de formules dont la conjonction est 
   quivalente  la formule de dpart. *)
let rec devl p = 
  let rec fn pos = function
  Et(p,p') as p'' -> if pos then fn pos p @ fn pos p' else [Non p'']
| Ou(p,p') as p'' -> if pos then [p''] else fn pos p @ fn pos p'
| Im(p,p') as p'' -> if pos then [p''] else fn (not pos) p @ fn pos p'
| Eq(p,p') as p'' -> if pos then fn pos (Et (Ou (Non p,p'), Ou (p,Non p')))
                     else fn pos (Ou (Et (p,p'), Et (Non p,Non p')))
| Non p -> fn (not pos) p
| Al p as p' -> if pos then map (fun x -> Al x) (fn pos p) else [Non p']
| Ex p as p'-> if pos then [p'] else map (fun x -> Al x) (fn pos p)
| Atom _ as p -> if pos then [p] else [Non p]
| Vrai -> if pos then [] else raise Fail
| Faux -> if pos then raise Fail else []
  in fn true p
;;


(* transformation d'un prdicat en un ensemble de clauses *)
let prop_clause p =
  let rec fn pos env = 
    let rec do_sk p = function 
      true -> let v = new_var () in fn pos (v::env) p
    | false -> let n = new_cst () in fn pos ((build_app n (uvars env p))::env) p

    and join (lp,ln) (lp',ln') = (union lp lp') ,(union ln ln')

    and uvars env f = (gn 0 p 
      where rec gn d = function
        Al p -> gn (d+1) p
      | Ex p -> gn (d+1) p
      | Atom t -> hn t
         where rec hn = (function
           BVar n -> if n < d then []
                     else let t = nth (n-d) env in
                     (match t.Val with Unknown _ -> [t] | _ -> [])
         | Fun (_,l) -> flat_map hn l)
      | Et (p,p') -> union (gn d p) (gn d p')
      | Ou (p,p') -> union (gn d p) (gn d p')
      | Im (p,p') -> union (gn d p) (gn d p')
      | Eq (p,p') -> union (gn d p) (gn d p')
      | Non p -> gn d p
      | Vrai | Faux -> [])

    and do_clause pos p pos' p' = function
      true -> union (fn pos env p) (fn pos' env p') 
    | false ->  (try let l = fn pos env p in
                    try let l' = fn pos' env p' in
                      it_list (fun res cl -> union res (map (join cl) l')) [] l 
                    with Proved -> l
                  with Proved -> fn pos' env p')

    in function
    Al p -> do_sk p pos 
  | Ex p -> do_sk p (not pos)
  | Atom t -> 
    let rec fn = function
      BVar n -> nth n env
    | Fun (n,l) -> build_app n (map fn l)
    in let t = fn t in if pos then [[t],[]] else [[],[t]]
  | Et (p,p') -> do_clause pos p pos p' pos
  | Ou (p,p') -> do_clause pos p pos p' (not pos)
  | Im (p,p') -> do_clause (not pos) p pos p' (not pos)     
  | Eq (p,p') -> if pos then 
                   fn pos env (Et (Ou (Non p,p'), Ou (p,Non p')))
                 else
                   fn pos env (Ou (Et (p,p'), Et (Non p,Non p')))
  | Non p -> fn (not pos) env p
  | Vrai -> if pos then [] else raise Proved  
  | Faux -> if pos then raise Proved else []

  in fn true [] p
;;

(* affichage d'une clause *)
let print_clause (n,_,_,l1,l2) =
  print_char `<`; print_int !n; print_char `>`;
  let g f (l1,l2) = do_list f l1; do_list f l2
  and h f (l1,l2) = 
    if l1 = [] & l2 = [] then print_string " []";   
    do_list (fun t -> print_string " ~"; f t) l2;
    do_list (fun t -> print_string " "; f t) l1
  in
  do_print g h (l1,l2);
  print_newline ()
;;

(* comparaison des infos d'une clause *)
let compare_info 
  (l1,l2,a1,a2,a3,a4,a5,a6) (l1',l2',a1',a2',a3',a4',a5',a6') =
  l1 > l1' or l2 > l2' or a1 > a1' or a2 > a2' or a3 > a3' or a4 > a4'
  or a5 > a5' or a6 > a6'
;;

(* doit-on afficher des informations pendant la rsolution *)
let verbose = ref false;;

(* Quelle stratgie *)
let strategie = ref Hyper;;

(* la liste des axiomes (conservs pour plusieurs rsolution). *)
let axiom = ref ([] : (Term list * Term list) list);;

(* la table des clauses, la clef entire donne la "gnration" de la clause *)
let clauses = (myhashtbl__new 23 myhashtbl__Standard myhashtbl__Standard 
                  : (int,clause list) myhashtbl__t);;

(* gnration courante (en train d'tre rsolue avec les gnrations 
   plus rcentes) *)
let cur_gen = ref 1;;

(* gnration maximum contenant au moins une clause *)
let max_gen = ref 0;;

(* fabrication des numros des clauses *)
let cur = ref 1;;
let new () = let r = !cur in cur := r + 1; ref r;;

(* fabrication des informations sur une liste de litraux *)
let linfo l = 
  it_list (fun (a1,a2,a3) t -> 
    let ni = get_info t in
    (a1+ni.length,max a2 ni.depth_max,max a3 ni.depth_min)) 
  (0,0,0) l
;;

(* gnralisation de map et do_list aux paires de listes *)
let map_pair fn (lp,ln) = 
  let fn = fun x -> snd (fn x) in
  (map fn lp, map fn ln);;
let do_pair fn (lp,ln) = do_list fn lp; do_list fn ln;;

(* fabrication d'une nouvelle clause (avec copie des termes !) *)
let build_clause ax lp ln =
  let (lp,ln) = copy map_pair do_pair (lp,ln) in
  let (a1,a2,a3) = linfo lp and (a4,a5,a6) = linfo ln in
  let info = (list_length lp,list_length ln,a1,a2,a3,a4,a5,a6) in
  (new (),ax,info,lp,ln)
;;

(* copie d'une clause*)
let copy_clause (r,ax,info,lp,ln) =
  let (lp,ln) = copy map_pair do_pair (lp,ln) in
  (r,ax,info,lp,ln)
;;

(* application d'une fonction  toutes les clauses *)
let it_clauses f a = 
  myhashtbl__it_table (fun a _ l -> it_list f a l) a clauses;;

(* test des tautologies *)
let tautologie lp ln = 
  exists (fun c -> exists (fun c' -> equal c c') ln) lp;;

(* algorithme de subsumption *)
let subsume (no,_,i,lp,ln) (no',_,i',lp',ln') =
  let rec fn = fun
    [] [] -> ()
  | [] (t::ln) -> hn t [] ln ln'
  | (t::lp) ln -> hn t lp ln lp'
  and hn t lp ln = function
    [] -> raise Fail
  | t'::l' -> try 
                let b, modif = pmatch t t' in
                try
                  fn lp ln; undo modif
                with Fail -> undo modif; if b then hn t lp ln l' else raise Fail
              with Unif_fail -> hn t lp ln l'
  in 
  if (compare_info i i') then false else
  try 
    fn lp ln; 
    if !verbose then begin
      print_string "clause ";
      print_int !no';
      print_string " subsumed by ";
      print_int !no;
      print_newline ()
    end;
    true
  with Fail -> false            
;;

(* un type d'arbre utilis pour la contraction *) 
type 'a tree = N of 'a * 'a tree ref list ref | D
;;

(* trois fonctions auxilliaire pour la contraction *)
let delete a b = 
  let rec fn t =
    if t == a then b:=!a
    else match t with
      ref D -> ()
    | ref (N (cl,ptr)) as p -> p:=D; do_list fn !ptr
  in fn b; if !b == !a then raise Exit
;;

let exists_in_tree t f = fn t
  where rec fn = function
    ref D -> false
  | ref (N (cl,ptr)) -> if f cl then true else exists fn !ptr
;;

let tree_to_list t = fn [] t
  where rec fn acc = function
    ref D -> acc
  | ref (N (cl,ptr)) -> it_list fn (cl::acc) !ptr
;;

(* la contraction *)
let contract (_,ax,_,_,_ as cl) =
  let tree = ref (N (cl, ref [])) in
  let to_do = ref [tree] in
  begin 
    while !to_do <> [] do
      match !to_do with
        [] -> ()
      | t::l -> to_do := l; contract_one t
    done; 
    tree_to_list tree
  end 

  where rec contract_one = function
    ref D -> ()
  | ref (N (no,_,_,lp,ln as cl, ptr)) as curptr ->

      let rec fn lp' ln' lp ln = 
        let rec hn t lp ln = function
          [] -> ()
        | t'::l' -> 
            (try 
              let modif = unif t t' in 
              let lp'',ln'' = copy map_pair do_pair (lp,ln) in
              undo modif;
              if tautologie lp'' ln'' then () else (
              let cl = build_clause ax lp'' ln'' in
              if !verbose then begin
                print_string "clause ";
                print_int !no;
                print_string " contract to ";
                print_clause cl
              end;
              if it_clauses (fun b c' -> b or (subsume c' cl)) false 
                then () else (
              if exists_in_tree tree (function c -> subsume c cl) 
                then () else (
              let ptrn = ref (N (cl, ref [])) in
              ptr:= ptrn :: !ptr;
              let rec kn = function
                ref D -> ()
              | ref (N (cl',ptr)) as ptro -> 
                  if subsume cl cl' then delete ptrn ptro
              in kn tree)))
            with Unif_fail -> ()); hn t lp ln l'
        in match lp,ln with
        [], [] -> ()
      | [], (t::ln) -> hn t lp' (ln'@ln) ln; fn lp' (t::ln') [] ln
      | (t::lp), ln -> hn t (lp'@lp) (ln'@ln) lp; fn (t::lp') ln' lp ln


    in try fn [] [] lp ln with Exit -> ();
    let ncl = !ptr in
    if ncl <> [] then to_do := !to_do @ ncl
;;

(* la rsolution, les litraux t et t'  rsoudre on t retirs des clauses 
   donnes en second et quatrime argument.
     teste pour les tautologies et la forward subsumption  *)
let rec resolve t (no,_,_,lp',ln') t' (no',_,_,lp'',ln'') = 
  if !no < 0 or !no' < 0  then [] else
  try 
    let modif = unif t t' in
    let lp = lp'@lp'' and ln = ln'@ln'' in
    let ((_,_,_,lp,ln) as cl) = build_clause false lp ln in
    undo modif;
    if !verbose then begin
      print_string "res ";
      print_int !no;
      print_char ` `;
      print_int !no';
      print_string " gives ";
      print_clause cl
    end;
    if lp = [] & ln = [] then raise Proved;
    if tautologie lp ln then [] else
    if it_clauses (fun b c' -> b or (subsume c' cl))
           false then []
    else [cl]
  with Unif_fail -> []
;;

(* ajout d'une clause dans la table des clauses.
      appel la contraction et effectue la backward subsumption *)
let add_clause nc =
  let nc = contract nc in
  myhashtbl__do_table (fun n l ->
    let nl = it_list (fun ng c -> 
    if (exists (fun c' -> subsume c' c) nc) then 
      (let (no,_,_,_,_) = c in no:= -1;ng) else c::ng) [] l in
    myhashtbl__remove clauses n;
    myhashtbl__add clauses n nl) clauses;
  do_list (fun ((no,_,(i,j,_,a2,a3,_,a5,a6),lp,ln) as c) ->
    if !verbose then begin
      print_string "keep clause "; print_int !no; print_newline ()
    end;
    let gen = (i + j) * (a2 + a5 + 1) + !cur_gen in
    if gen > !max_gen then max_gen := gen;
    let old = 
      try 
        let l = myhashtbl__find clauses gen in
        myhashtbl__remove clauses gen; l
      with Not_found -> []
    in myhashtbl__add clauses gen (c::old)) nc
;;

(* teste si une clause est unitaire *)
let unit (_,_,_,lp,ln) = (fun
  [] [_] -> true
| [_] [] -> true
| _ _ -> false) lp ln
;;

(* teste suivant la stratgie si une clause doit tre ajoute.
   Dans le cas des set-of-support strategy, les clauses unitaires sont 
   toujours ajoutes. *)
let rec add_new = function
  [] -> []
| (no,ax,i,lp,ln as cl)::l ->
    if !strategie = Binary or (!strategie = Hyper & ln = [])
                           or (!strategie = NegHyper & lp = []) then begin
      add_clause cl; 
      add_new l
    end else begin
      if unit cl then add_clause cl;
      cl::(add_new l)
    end
;;

(* table indiquant le nombre de possibilit d'unification d'un
   litral.  Cela permet de ne pas calculer plusieurs fois cette
   valeur. Cette table est efface  chaque nouvelle gnration *)
let pondt = myhashtbl__new 401 (myhashtbl__Custom key) 
                               (myhashtbl__Custom equal);;

(* fonction recherchant l'lment minimum d'une liste *)
let min_first cmp l =
  let rec fn m acc = function 
    x::l -> if cmp x m then fn x (m::acc) l else fn m (x::acc) l
  | [] -> m::acc
  in match l with
    [] -> []
  | x::l -> fn x [] l
;;

(* trie les litraux d'une liste pour placer en premier le litral 
   ayant le moins d'unification possible *)
let sort_lit  l =
  let pondere lit = 
    let c = ref 0 in
    let rec fn = function
      [] -> ()
    | t::l ->
      (try let modif = unif t lit in c:=!c+1; undo modif
      with Unif_fail -> ());
      fn l
    in
    let gn (no,_,_,lp,ln) = if !no > 0 then (
                           if !strategie = Hyper then 
                             if ln = [] then fn lp else () 
                           else if !strategie = NegHyper then 
                             if lp = [] then fn ln else ()
                           else ())
    in
    for i = 1 to !cur_gen do
      try 
        let cur = myhashtbl__find clauses i in
        do_list gn cur
      with Not_found -> ()
    done;
    !c
  in
  let hpond lit = myhashtbl__add_find pondt lit pondere in
  let l = map (fun x -> hpond x, x) l in
  let l = min_first (fun x y -> fst x <= fst y) l in
  if fst (hd l) = 0 then raise Exit;
  map snd l
;;

(* calcule toutes les rsolutions possible de deux clauses suivant la
   stratgie.  le boolen donn en premier argument permet de
   distinguer l'appel initial des appels rcursifs (inutile pour la
   rsolution binaire). *)
let rec resolve_all ft (no,ax,i,lp,ln) (no',ax',i',lp',ln') =

  if !no > 0 & !no' > 0 &
     ((!strategie = Binary) or 
     (!strategie = Hyper & ln = []) or 
     (!strategie = NegHyper & lp' = [])) then

  let rec fn lp0 = function
    [] -> ()
  | p::lp -> gn p (lp0@lp) [] ln'; 
             if !strategie <> NegHyper or ft then fn (p::lp0) lp
  and gn p lp ln0' = function
    [] -> ()
  | n::ln' -> 
    let nc = resolve p (no,ax,i,lp,ln) n (no',ax',i',lp',(ln0'@ln')) in 
    let nc = add_new nc in 
    let rec kn acc = function
      [] -> acc	
    | (no,ax,i,lp,ln as cl)::l ->
        try match !strategie with
          Binary -> failwith "bug in resolve all"
        | Hyper -> if list_length ln > 1  
                   then kn ((no,ax,i,lp,sort_lit ln)::acc) l
                   else kn (cl::acc) l
        | NegHyper -> 
                   if list_length lp > 1
                   then kn ((no,ax,i,sort_lit lp,ln)::acc) l
                   else kn (cl::acc) l
        with Exit -> kn acc l
    in
    let nc = kn [] nc in
    if nc <> [] then myhashtbl__do_table 
      (fun g cur -> if g <= !cur_gen then do_next_sgen false nc cur) 
      clauses;
    if !strategie <> Hyper or ft then gn p lp (n::ln0') ln'
  
  in fn [] lp

(* Fonction testant toutes les rsolutions possibles  l'intrieur
   d'une liste de clauses *)
and do_next_gen = fun
  [] -> ()
| (cur::l) ->
     resolve_all true cur (copy_clause cur);
     do_list (fun cur' -> resolve_all true cur cur';
                          resolve_all true cur' cur) l;
     do_next_gen l 

(* Fonction testant toutes les rsolutions possibles entre deux listes
   de clauses *)
and do_next_sgen ft = fun
  [] _ -> ()
| (cur::l) l' ->
     do_list (fun cur' -> resolve_all ft cur cur';
                          resolve_all ft cur' cur) l';
     do_next_sgen ft l l'
;;

(* fonction d'initialisation *)
let do_init typ = myhashtbl__clear clauses; cur := 1;
                 cur_gen := 0; max_gen := 0;
                 strategie := typ;
                 reset_ref ();
                 do_list (function lp,ln -> 
                          if not tautologie lp ln then begin
                            let cl = build_clause true lp ln in
                            if !verbose then print_clause cl;
                            if not (it_clauses 
                                      (fun b c' -> b or (subsume c' cl)) 
                                       false)
                            then add_clause cl
                            end) !axiom

;; 

(* fonction ajoutant un axiome *)
let addax prop =
  let first = prop_clause prop in 
  axiom := first@(!axiom)
;;


(* fonction effectuant la rsolution (le premier argument tant la
   stratgie et le second la formule  prouver) *)
let oneres typ prop =
  do_init typ;
  let first = prop_clause (Non prop) in 
  do_list (function lp,ln -> 
             if not tautologie lp ln then begin
               let cl = build_clause true lp ln in
               if !verbose then print_clause cl;
               if not (it_clauses 
                         (fun b c' -> b or (subsume c' cl)) 
                         false)
               then add_clause cl
               end) first;
  cur_gen := 1;
  try 
    while true do
      if !strategie <> Binary then myhashtbl__clear pondt;
      (try let new = myhashtbl__find clauses !cur_gen in
      if !verbose then print_endline "starting new generation.";
      let rec fn n = 
        if n = !cur_gen then do_next_gen new
        else (
          (try let old = myhashtbl__find clauses n in
            do_next_sgen true new old
          with Not_found -> ());
          fn (n+1)) in
      fn 1
      with Not_found -> ());
      cur_gen:=!cur_gen+1;
      if !cur_gen > !max_gen then raise Fail
    done; true
  with Proved -> true | Fail -> false
;;

(* fonction appelant la fonction "devl" puis la rsolution sur chacune des
   formules *) 
let res typ prop =
  for_all (oneres typ) (devl prop)
;;




