(* Type usuel rien ou objet *)
type 'a obj = Nothing | Something of 'a;;

(* Identificateurs des variables (utile uniquement pour l'affichage) *)
type var_info == int;;

(* Identificateurs des constantes (on distingue les prdicats et les fonctions) 
     Les info. dans cst_info ne sont utiliss que pour l'affichage et le 
     parsing. On utilise l'galit physique pour comparer les symboles. *)
type cst_type = Pred_sym | Fun_sym;;
type cst_info = {Cst_name : string; Cst_arity : int; Cst_type : cst_type};;

(* Information sur les termes pour optimiser le pattern-matching *)
type Term_info = {mutable length : int;
                  mutable depth_max : int;
                  mutable depth_min : int};;                  

(* Representation des termes comme des graphes *)
type Term = {mutable Val : Value; 
                mutable Dp : bool; 
                mutable Ad : bool; (* Dp et Ad utiliss pour le test 
                                   d'acyclicite *)
                mutable Copy : (Term_info * Term) obj (* Pour la copy *)}

and Value =
  Cst of (cst_info * Term vect * Term_info) 
    (* Une constante applique  ces arguments *)
| Link of Term
    (* Le lien utilis pour l'unification et le pattern-matching et 
       exprimant que deux sous graphes doivent tre gaux.*)
| Unknown of var_info
    (* Les variables *)
;;

(* L'exception leve en cas d'chec *)
exception Unif_fail;;

(* Le plus grand entier en Caml-Light *)
let max_int = 65536 * 16384 - 1;;

(* Construction du champs Term *)
let build v = {Val = v; Dp = false; Ad = false; Copy = Nothing};;

(* Gnration de nouvelles variables *)
let count_var  = ref 0;;
let new_var () = let r = !count_var in count_var:=r+1; build (Unknown r);;

(* Un objet quelconque pour l'initialisation des vecteurs *)
let dummy = build (Unknown (-1));;

(* Suivi des liens et raccourcicement du chemin (la fonction G) *)
let rec getv modif t = fn t 
  where rec fn t = match t.Val with
    Cst _ | Unknown _ -> t
  | Link t' -> 
      let t'' = fn t' in
      if t' != t'' then (modif:=(t,t.Val)::!modif; t.Val <- Link t''); t''
;;

(* Suivi des liens sans raccourcicement du chemin (quand cela reste
   linaire, il faut mieux utiliser getvs qui est "propre" *)
let rec getvs t = match t.Val with
  Cst _ | Unknown _ -> t
| Link t' -> getvs t'
;;
 
(* Gnralisation de la fonction do_vect *)
let do_vect2 f v1 v2 =
  let n = vect_length v1 in
  for i = 0 to (n-1) do f v1.(i) v2.(i) done
;;

(* Fonction effectuant une liste de modifications pour "dfaire" les
   modifications faites par unif, pmatch ou equal *)
let undo modif = do_list (function t,v -> t.Val <- v) modif;;

(* L'unification de termes rationnels. En cas d'echec les modifications 
   sont "dfaites" et l'exception Unif_fail est leve. En cas de russite
   la liste des modifications est renvoye. *)
let rec punif t1 t2 = 
  let modif = ref [] in
  (try fn t1 t2; !modif
  with e -> undo !modif; raise e)
  where rec fn t1 t2 = 
    let t1 = getv modif t1 and t2 = getv modif t2 in
    if t1 == t2 then () else
    match (t1.Val,t2.Val) with
      Cst (n1,v1,_), Cst (n2,v2,_) -> 
        if n1 == n2 then begin
          modif:=(t1,t1.Val)::!modif;
          t1.Val <- Link t2; do_vect2 fn v1 v2 
        end else
          raise Unif_fail
    | Unknown _, (Cst _ | Unknown _) ->
        modif:=(t1,t1.Val)::!modif;
        t1.Val <- Link t2
    | Cst _, Unknown _ ->
        modif:=(t2,t2.Val)::!modif;
        t2.Val <- Link t1
    | _, _ -> raise (Failure "bug: punif got link !") 
;;

(* Test d'acyclicit avec nettoyage des champs Dp et Ad en cas de
   succs ou d'chec. Renvoie () : unit en cas de succs et lve
   Unif_fail sinon.  Le champs Ad est  true pour les variables qui
   sont sur le chemin de la racine  la position courante, tandis que
   Dp est  vrai pour les autres variables dj testes. *)
let finite t = (try fn t; clean t with e -> clean t; raise e)
where rec fn t = 
  let t = getvs t in
  if t.Ad then raise Unif_fail;
  if t.Dp then () else
    match t.Val with 
      Cst(_,v,_) -> 
        t.Ad <- true;
        do_vect fn v;  
        t.Ad <- false;
        t.Dp <- true
    | _ -> 
        t.Dp <- true
and clean t =
  let t = getvs t in
  if t.Dp or t.Ad then begin
    t.Dp <- false; t.Ad <- false;
    match t.Val with 
      Cst(_,v,_) -> 
        do_vect clean v
    | _ -> ()
  end
;;

(* L'unification complte elle-mme. En cas d'echec les modifications 
   sont "dfaites" et l'exception Unif_fail est leve. En cas de russite
   la liste des modifications est renvoye. *)
let unif t t' =
  let modif = punif t t' in 
  try
    finite t; modif
  with e -> undo modif; raise e
;;

(* Suivi des liens et raccourcicement du chemin, de plus on renvoie 
   un boolen pour dire si on a travers un lien (utilis par le pattern
   matching, car aprs avoir travers un lien il est interdit de substituer
   les variables) *)
let getv' modif t =
  match t.Val with
    Cst _ | Unknown _ -> t, false
  | Link t' -> 
      let t'' = getv modif t' in
      if t' != t'' then (modif:=(t,t.Val)::!modif; t.Val <- Link t''); 
      (t'',true)
;;

(* Test de pattern-matching et d'galit. 
   Pour pmatch : En cas d'echec les modifications sont "dfaites" et 
     l'exception Unif_fail est leve. En cas de russite la liste des 
     modifications est renvoye. 
   Pour equal : Toujours propre, et renvoie un boolen *)
let pmatch,equal = 
  let b = ref false in
  let modif = ref [] in
  let rec fn lock t1 t2 = 
  let (t1,lock1) = getv' modif t1 and (t2,_) = getv' modif t2 in
  let lock = lock or lock1 in
  if t1 == t2 then () else
  match (t1.Val,t2.Val) with
    Cst (n1,v1,i1), Cst (n2,v2,i2) ->
      if n1 != n2 or 
         i1.length > i2.length or 
         i1.depth_max > i2.depth_max or
         i1.depth_min > i2.depth_min then
           raise Unif_fail;
      modif:=(t1,t1.Val)::!modif;
      t1.Val <- Link t2;
      do_vect2 (fn lock) v1 v2 
  | Cst _, Unknown _ ->
      raise Unif_fail
  | Unknown _, _ ->
      if lock then raise Unif_fail;
      modif:=(t1,t1.Val)::!modif;
      t1.Val <- Link t2;
      b:= true
  | _, _ -> raise (Failure "bug: pmatch got link !") 
  in (fun t1 t2 -> b:= false; modif := [];
                   try fn false t1 t2; (!b,!modif)
                   with e -> undo !modif; raise e), 
     (fun t1 t2 -> b:= false; modif := [];
                   try fn true t1 t2; undo !modif; true 
                   with Unif_fail -> undo !modif; false )
;;

(* fonction de copie d'une structure contenant des termes (les
   fonctions fmap et fdo doivent tre l'analogue de map et do_list
   pour la structure concerne). La copie conserve le partage et
   nettoie le champs Copy. Elle marche aussi sur des termes
   cycliques. Les informations Term_info sont recalcules (car
   l'argument provient en gnral du rsultat d'une unification ou du
   pattern-matching et ces informations ne sont en gnral plus
   valable *)

let copy fmap fdo obj = (
  let r = fmap fn obj in 
  (fdo clean obj : unit); r)
  where rec fn t = let t = getvs t in
    match t.Copy with
      Something c -> c
    | Nothing ->
      match t.Val with
        Cst(f,v,_) -> 
          let ni = {length = 1; depth_max = 0; 
                    depth_min = if vect_length v = 0 then 0 else max_int} in
          let nv = make_vect (vect_length v) dummy in
          (* La ligne suivante est  cette place pour autoriser la copie
             de termes cycliques *)
          let c = ni, build (Cst(f,nv,ni)) in t.Copy <- Something c;
          for i = 0 to (vect_length v) - 1 do
            let (ni',nt') = fn v.(i) in
            nv.(i) <- nt';
            ni.length <- ni.length + ni'.length;
            ni.depth_max <- max ni.depth_max (1 + ni'.depth_max);
            ni.depth_min <- min ni.depth_min (1 + ni'.depth_min)
          done;
          c
      | Unknown _ -> 
          let ni = {length = 1; depth_max = 0; depth_min = 0} in
          let c = ni, new_var () in t.Copy <- Something c; c
      | _ -> failwith "Bug: copy got link"
  and clean t = 
  let t = getvs t in
  if t.Copy = Nothing then () else begin
    t.Copy <- Nothing;
    match t.Val with 
      Cst(_,v,_) -> 
        do_vect clean v
    | _ -> ()
  end
;;

(* Les fonctions qui suivent sont utiles, mais n'ont pas de rapports
   directs avec l'unification. *)

(* fonction renvoyant une clef de hashage pour un terme. Cette clef
   satisfait : si equal t u = true alors key t = key u. On ne peut pas
   utiliser la fonction de hashage polymorphe de Caml-Light pour deux
   raisons : 
     - La structure est cyclique.  
     - L'galit qui nous intresse est "au nom des variables prs".*)

let rec key t =
  let hashval = ref 0 in
  let dummy = {length = 0; depth_max = 0; depth_min = 0} in
  let combine_small x = hashval := abs (!hashval * 7 + x)
  and combine_big x = hashval := abs (!hashval * 65599 + x) in
  let to_clean = ref [] in
  let nv = ref 0 in
  let nm = ref 75 in
  let rec fn p t = 
      let t = getvs t in
      match t.Val with
        Cst(f,v,_) -> 
          combine_small 0;
	  combine_big (hashtbl__hash f.Cst_name);
          nm:=!nm-1;
          if !nm > 0 & p > 0 then do_vect (fn (p - 1)) v
      | Unknown n-> 
          nm:=!nm-1;
          let k = match t.Copy with
            Nothing ->
              nv:=!nv+1;
              t.Copy <- Something(dummy, build (Unknown !nv));
              to_clean:= t::!to_clean;
              !nv
          | Something(_, {Val = Unknown k}) -> k
          | _ -> failwith "bug in key"
          in
          combine_small 1;
          combine_small k
      | _ -> failwith "Bug: key got link"
  in fn 10 t;
  do_list (fun t -> t.Copy <- Nothing) !to_clean;
  !hashval
;;

(* Affichage d'un terme, en utilisant le partage pour pouvoir afficher
   des gros termes ! La fonction do_print est un analogue fonctionnel
   (cf copie) permettant d'afficher une structure contenant des termes
   en visialisant le partage entre les diffrents termes. *)

let print_term, do_print =
  let cur = ref 1 in
  let dummy = {length = 0; depth_max = 0; depth_min = 0} in
  let new_tag () = let r = !cur in 
    cur:=r+1;  
    Something (dummy, build (Unknown r))
  and get_tag t = match t.Copy with
    Nothing -> raise Not_found
  | Something (_,{Val = Unknown r}) -> "#"^(string_of_int r)
  | _ -> raise (Failure "bug 3 in print_term")
  in
  let rec fn t = 
    let t = getvs t in
    match t.Val with
        Unknown r -> print_string "?"; print_string (string_of_int r)
      | Cst ({Cst_name = s},v,_) ->
          if v = [||] then print_string s else
          begin
            try 
              let tag = get_tag t in
              print_string tag;
              if t.Dp then begin
                t.Dp <- false; 
                print_string "=";
                raise Not_found
              end
            with Not_found ->
              print_string "(";print_string s;
              do_vect (fun t -> print_string " "; fn t) v;
              print_string ")"
          end
      | _ -> raise (Failure "bug 1 in print_term")
  and look_for_share t =
    let t = getvs t in
    match t.Val with
        Unknown _ -> ()
      | Cst ({Cst_name = s},v,_) ->
          if v = [||] then () else
          if t.Dp then t.Copy <- new_tag ()
          else begin
            t.Dp <- true; do_vect look_for_share v
          end
      | _ -> raise (Failure "bug 2 in print_term")
  and clean t = 
    let t = getvs t in
    if t.Copy = Nothing & not t.Dp then () else begin
      t.Copy <- Nothing;
      t.Dp <- false;
      match t.Val with 
        Cst(_,v,_) -> 
          do_vect clean v
      | _ -> ()
    end
  in
  (fun t -> cur:=1; look_for_share t; fn t; clean t),
  (fun (g : (Term -> unit) -> Term list * Term list -> unit) 
       (h : (Term -> unit) -> Term list * Term list -> unit) o -> 
     cur:=1; g look_for_share o; h fn o; g clean o)
;;                 

(* Lecture des informations sur les termes *)
let get_info t = match (getvs t).Val with
  Unknown _ -> {length = 1; depth_max = 0; depth_min = 0}
| Cst(_,_,i) -> i
| _ -> failwith "Bug in get_info"
;;

(* Construction d'une application avec calcul des informations
   (utilis par le parser pour construire des termes. *)
let build_app f l = 
  let len = list_length l in
  let nv = make_vect len dummy in
  let ni = {length = 1; depth_max = 0; 
            depth_min = if len = 0 then 0 else max_int} in
  let i = ref 0 in
  do_list (
    fun t -> 
      let ni' = get_info t in
      ni.length <- ni.length + ni'.length;
      ni.depth_max <- max ni.depth_max (1 + ni'.depth_max);
      ni.depth_min <- min ni.depth_min (1 + ni'.depth_min);
      nv.(!i) <- t;
      i:=!i+1) 
    l;
  build (Cst(f,nv,ni))
;;
