(* type dcrivant les if-expressions *)

type if_expr =
  iVrai | iFaux | iVar of int
| If of if_expr * if_expr * if_expr
;;

(* form_to_normal_if p fabrique une if-expression normale quivalente
    la formule p *)

let form_to_normal_if p = 
  (* sachant que 
       - tbl est une valuation partielle associant un boolen  
         certaines variables propositionnelles
       - l est une liste de paires de formules
     fn tbl l p, fabrique une if-expression quivalente  la formule 
     IF(p,l) o les variables dclares dans tbl substitues par leur valeur.  
     IF(p,l) tant dfinit par 
       - IF(p,[]) = p
       - IF[p,(p1,q1)::l] = if p then IF[p1,l] else IF[q1,l] 
  *)
  let rec fn tbl l = function
    Vrai -> gn tbl true l
  | Faux -> gn tbl false l
  | Var n -> (
     try gn tbl (assoc n tbl) l 
     with Not_found -> If (iVar n, gn ((n,true)::tbl) true l,
                                  gn ((n,false)::tbl) false l))
  | Bin (Et,p1,p2) -> fn tbl ((p2,Faux)::l) p1
  | Bin (Ou,p1,p2) -> fn tbl ((Vrai,p2)::l) p1
  | Bin (Im,p1,p2) -> fn tbl ((p2,Vrai)::l) p1
  | Bin (Eq,p1,p2) -> fn tbl ((p2,Non p2)::l) p1
  | Non (Non p1) -> fn tbl l p1
  | Non p1 -> fn tbl ((Faux,Vrai)::l) p1

  and gn tbl b = function
    [] -> if b then iVrai else iFaux 
  | (p1,p2)::l' -> if b then fn tbl l' p1 
                        else fn tbl l' p2 

  in fn [] [] p
;;

(* taut_if p vrifie que p est une tautologie, en utilisant le mme
algorithme que la fonction prcdente, mais sans calculer la
if-expression. La fonction vrifie simplement que si la if-expression
avait t construite, toutes ses feuilles seraient gales  Vrai (ce
qui pour une if-expression normale est quivalent  dire qu'il s'agit
d'une tautologie *)

let taut_if p =
  let rec fn tbl l = function
    Vrai -> gn tbl true l
  | Faux -> gn tbl false l
  | Var n -> (
     try gn tbl (assoc n tbl) l 
     with Not_found -> gn ((n,true)::tbl) true l & gn ((n,false)::tbl) false l)
  | Bin (Et,p1,p2) -> fn tbl ((p2,Faux)::l) p1
  | Bin (Ou,p1,p2) -> fn tbl ((Vrai,p2)::l) p1
  | Bin (Im,p1,p2) -> fn tbl ((p2,Vrai)::l) p1
  | Bin (Eq,p1,p2) -> fn tbl ((p2,Non p2)::l) p1
  | Non (Non p1) -> fn tbl l p1
  | Non p1 -> fn tbl ((Faux,Vrai)::l) p1

  and gn tbl b = function
    [] -> b
  | (p1,p2)::l' -> if b then fn tbl l' p1 
                        else fn tbl l' p2 

  in fn [] [] p
;;

(* simplify tbl p simplifie une formule p sachant que certaines
variables sont gales  Vrai ou Faux, comme indiqu dans tbl.
tbl est un tableau d'objets du type val dfini ci dessous. *)

type val = 
  ValVrai
| ValFaux
| NoVal
| NoOcc;;

let simplify tbl p = 
  let rec fn = function
    Vrai | Faux as p -> p
  | Var n as p -> (match tbl.(n) with
                   ValVrai -> Vrai
                 | ValFaux -> Faux
                 | _ -> p)
  | Non p as pt -> (match fn p with Vrai -> Faux | Faux -> Vrai | _ -> pt)
  | Bin(c,p1,p2) as p -> match (c,fn p1,fn p2) with
      (Et,Vrai,p2) -> p2
    | (Et,p1,Vrai) -> p1
    | (Et,Faux,_)  -> Faux
    | (Et,_,Faux)  -> Faux

    | (Ou,Faux,p2) -> p2
    | (Ou,p1,Faux) -> p1
    | (Ou,Vrai,_)  -> Vrai
    | (Ou,_,Vrai)  -> Vrai

    | (Im,Faux,p2) -> Vrai
    | (Im,Vrai,Faux) -> Faux
    | (Im,p1,Faux) -> Non p1
    | (Im,Vrai,p2)  -> p2
    | (Im,p1,Vrai)  -> Vrai

    | (Eq,Vrai,p2) -> p2
    | (Eq,p1,Vrai) -> p1
    | (Eq,Faux,Faux) -> Vrai
    | (Eq,p1,Faux) -> Non p1
    | (Eq,Faux,p2) -> Non p2

    | (_,p1',p2') as p' -> if p1 == p1' & p2 == p2' then p else Bin p'
  in fn p
;;

(* type permettant de dire si une variable occurre positivement,
   ngativement ou les deux dans une formule *)

type occ_typ = Neg | Pos | Both;;

(* action de la ngation sur le type d'occurrence *)

let neg_occ = function
  Neg -> Pos
| Pos -> Neg
| Both -> Both
;;

(* action de la conjonction ou de la disjonction sur le type d'occurrence *)

let merge_occ = fun
  Neg Neg -> Neg
| Pos Pos -> Pos
| _ _ -> Both
;;

(* type dcrivant des formules :
   - Bool true   dcrit une formule vrai
   - Bool false  dcrit une formule fausse
   - Form l dcrit une formule ayant au moins une variable libre. 
   l est une liste de couples (v,t) o v est une des variables 
   libres de la formule et t le type d'occurrence de cette variable.
   Les variables libres de la formule sont toutes prsentes exactement
   une fois dans l et ordonnes en ordre croissant *)

type mini_prop =
  Bool of bool
| Form of (var*occ_typ) list
;;

(* union_sort b l l' : calcule l'union des deux listes de variables l
et l' en supposant que l et l' sont tries en ordre croissant. le
rsultat tant lui aussi tri. Les informations correspondant  la
positivit des occurrences sont regroups en utilisant merge_occ *)

let rec union_sort = fun
  [] l -> l
| l [] -> l
| ((n,t as c)::l as l0) ((n',t' as c')::l' as l0') -> 
    if n < n' then c::(union_sort l l0') 
    else if n > n' then c'::(union_sort l0 l')
    else (n, merge_occ t t')::(union_sort l l')
;;

(* similaire  union_sort, mais les informations de type sont fixes a
Both (utilis pour l'quivalence) *)

let rec eq_union_sort = fun
  [] l -> l
| l [] -> l
| ((n,t as c)::l as l0) ((n',t' as c')::l' as l0') -> 
    if n < n' then c::(union_sort l l0') 
    else if n > n' then c'::(union_sort l0 l')
    else (n, Both)::(union_sort l l')
;;

(* action de la conjonction sur deux mini-formules *)

let mand = fun
  (Bool false as p) _ -> p
| _ (Bool false as p) -> p
| (Bool true as p) (Bool true) -> p
| (Bool _) (Form l as p) -> p
| (Form l as p) (Bool _) -> p
| (Form l) (Form l') -> Form (union_sort l l')
;;

(* action de la disjonction sur deux mini-formules *)

let mor = fun
  (Bool true as p) _ -> p
| _ (Bool true as p) -> p
| (Bool false as p) (Bool false) -> p
| (Bool _) (Form l as p) -> p
| (Form l as p) (Bool _) -> p
| (Form l) (Form l') -> Form (union_sort l l')
;;

(* action de la ngation sur une mini-formule *)

let mnot = function
  Bool b -> Bool (not b)
| Form l -> Form (map (function n,t -> n, neg_occ t) l)
;;

(* action de l'implication sur deux mini-formules *)

let mim p q = mor (mnot p) q
;;

(* action de l'quivalence sur deux mini-formules *)

let meq = fun
| (Bool b) (Bool b') -> Bool (b = b')
| (Bool b) (Form l as p) -> if b then p else mnot p
| (Form l as p) (Bool b) -> if b then p else mnot p
| (Form l) (Form l') -> Form (eq_union_sort l l')
;;

(* calcule la mini_prop dcrivant une formule *)

let rec prop_to_miniprop = function
  Var n -> Form [n,Pos]
| Vrai -> Bool true
| Faux -> Bool false
| Non f -> mnot (prop_to_miniprop f)
| Bin (c,f,f') -> 
      (match c with Et -> mand | Ou -> mor | Im -> mim | Eq -> meq)
      (prop_to_miniprop f) (prop_to_miniprop f')
;;

(* renvoie la variable d'indice maximum apparaissant dans f *)

let max_var f =
  let rec max cur = function
     Var n -> if n < cur then cur else n
   | Vrai | Faux -> cur
   | Non f -> max cur f
   | Bin(_,p1,p2) -> max (max cur p1) p2
  in max (-1) f
;;

(* construit une valuation : si une variable X occurre 
   positivement (resp. ngativement) dans F, F est une 
   tautologie si et seulement si F[X/Faux] (resp. Vrai)
   en est une aussi. *)
   
let rec mkval max_var l = 
  let tbl = make_vect (max_var+1) NoVal in 
  let trv = ref false in
  let rec fn = function
  [] -> ()
| (n,Pos)::l -> tbl.(n) <- ValFaux; trv := true; fn l
| (n,Neg)::l -> tbl.(n) <- ValVrai; trv := true; fn l
| (n,Both)::l -> fn l
  in fn l; tbl, !trv
;;

(* simplifie une formule en utilisant le principe ci-dessus jusqu' ce
   que toutes les variables libres occurrent  la fois positivement et
   ngativement. Attention, la formule rsultante n'est PAS
   quivalente  la formule de dpart. On ne peut donc pas utiliser
   cette simplification pour calculer une if-expression quivalente 
   une formule donne. La formule simplifie ainsi que le nombre de
   variables distinctes y apparaissant sont retourns *)

let rec rec_simplify max_var f =
  let mp = prop_to_miniprop f in
  match mp with
    Bool true -> Vrai, 0
  | Bool false -> Faux, 0
  | Form l -> 
      let val, b = mkval max_var l in
      if b then rec_simplify max_var (simplify val f) else f, list_length l
;;

(* calcule quelles variables occurrent dans f. Le rsultat est plac
dans un tableau en utilisant le constructeur NoOcc pour les variables
sans occurrence et le constructeur NoVal pour les autres *)

let vars max_var f =
  let vars = make_vect (max_var + 1) NoOcc in
  let rec set_vars = function
     Var n -> vars.(n) <- NoVal 
   | Vrai | Faux -> ()
   | Non f -> set_vars f
   | Bin(_,p1,p2) -> set_vars p1; set_vars p2
  in set_vars f; vars
;;

(* choisit la "meilleur variable". C'est  dire la variable X pour la
   quelle si l'on considere nn = nbr de vars dans F[X<-Vrai] aprs
   simplification et np = nbr de vars dans F[X<-Vrai] aprs
   simplification, alors 2^nn + 2^np est minimum. On utilise le fait
   que log_2 (2^nn + 2^np) = max nn np + log_2 (1 + 4^-|nn-np|) afin
   d'viter les overflows si l'on a plus de mille variables 
*)

let best_choice max_var f =
  let tbl = vars max_var f in
  let best = ref (Faux,Faux) in
  let bval = ref (float_of_int (2 + max_var)) in
  for i = 0 to max_var do
    if tbl.(i) <> NoOcc then begin
      tbl.(i) <- ValVrai;
      let fp,np = rec_simplify max_var (simplify tbl f) in
      tbl.(i) <- ValFaux;
      let fn,nn = rec_simplify max_var (simplify tbl f) in
      tbl.(i) <- NoVal;
      let n' = (float_of_int (max nn np))
               +. log (1. +. power 4. (-. (float_of_int (abs(nn-np))))) 
                  /. log 2. in 
      if n' <. !bval then (best := fp,fn; bval := n')
    end
  done;
  !best
;;

(* La fonction vrifiant si f est une tautologie ... *)

let rec taut_good_if f = 
  let max_var = max_var f in
  let rec fn = function
      Vrai -> true
    | Faux -> false
    | f -> 
        let f,f' = best_choice max_var f in
        fn f & fn f'
  in fn f
;;

(* cet algorithme peut tre optimiser car l'on parcours trop souvent
la formule. On peut grouper certains calculs et conserver des
informations  chaque noeud de la formule (tel que la mini_prop
associe au noeud, car l'on fait de nombreuse fois le mme calcul
lorsqu'une sous-formule reste inchange pendant quelques temps). Cela
devrait faire gagner un facteur constant assez important. *)

(* test sur une Sparc 10 avec des ensembles de clauses de taille 3
gnres alatoirement (cf appendix).

20 variables 150 clauses de taille 3
#let f = random_form 20 150 3;;
#chrono taut_good_if f;;
Time = 78.216667s of S+U time
- : bool = true
#chrono taut_if f;;
Time = 59.083333s of S+U time
- : bool = true
pas encore assez de variables ...

30 variables 150 clauses de taille 3
#let f = random_form 30 150 3;;
#chrono taut_good_if f;;
Time = 386.750000s of S+U time
- : bool = true
#chrono taut_if f;;
Time > 1200s of real time

40 variables 150 clauses de taille 3
#let f = random_form 40 150 3;;
#chrono taut_good_if f;;
Time = 82.9666666667s of S+U time
- : bool = false
#chrono taut_if f;;
Time > 1200s of real time
*)
